#!/usr/bin/perl
#
# SCoRE - Search a Corpus for Regular Expressions
#
# Perl module for BNC-specific routines
#
# (c) Matthew Purver, King's College London, 2001
#
# $Log: BNC.pm,v $
# Revision 1.1  2001/11/09 12:30:59  purver
# Initial revision
#

package BNC;

# export subroutines
require Exporter;
our @ISA = qw( Exporter );
our @EXPORT = qw( checkHeaderType 
		  displayOptions
		  getHeaderInfo
		  getSentence
		  initBuffer
		  noTags 
		  skipHeader 
		  skipAlign 
		  stringToRegExp 
		  uttSep );


# sentence/turn separator info
sub uttSep {
    my( $turns ) = $_[0];

    $uttSep = '<s [^<>]*>';
    $uttPat = '<s \[\^<>\]\*>';
    if ( $turns ) {
	$uttSep =~ s/<s/<u/;
	$uttPat =~ s/<s/<u/;
    }

    return ( $uttSep, $uttPat );
}


# share default options between this module and scripts
sub displayOptions {
    $sentNum = $_[0];
    die "sentNum must be defined\n" unless defined( $sentNum );

    $startMark = '__B__';
    $endMark = '__E__';

    return ( $startMark, $endMark );
}


# empty line buffer
sub initBuffer {
    $lineBuffer = '';
}


# discard to end of <teiHeader> section (or fail)
sub skipHeader {
    while ( $lineBuffer = <STDIN> ) {
	chomp( $lineBuffer );
	return 1 if ( $lineBuffer =~ s/^.*\/teiHeader>// );
    }
    return 0;
}


# discard to end of <align> section (or fail)
sub skipAlign {
    while ( $lineBuffer = <STDIN> ) {
	chomp( $lineBuffer );
	return 1 if ( $lineBuffer =~ s/^.*\/align>// );
    }
    return 0;
}


# check <catRef target=> types suitable (or fail), and return <locale> & <activity> info
# if available
sub checkHeaderType {
    my ( $arg, @inc, @exc, $info, $iDesc, $success );
    my @args = @_;
    my $maxDesc = 3;

    # build list of include/exclude values
    foreach $arg (@args) {
	push( @inc, $arg ) if ( $arg =~ s/^\+// );
	push( @exc, $arg ) if ( $arg =~ s/^-// );
    }

    # get whole <teiHeader> section
    while ( $lineBuffer .= <STDIN> ) {
	chomp( $lineBuffer );
	last if ( $lineBuffer =~ /\/teiHeader>/ );
    }

    # get <catRef> target attribute
    if ( $lineBuffer =~ /<catRef target=\"([^\"]*)\"/ ) {
	# split into array of values
	@chk = split( ' ', $1 );
	foreach $chk (@chk) {
	    # fail if any exclude values match
	    foreach $exc (@exc) {
		return 0 if $exc eq $chk;
	    }
	    # succeed if only one include value matches
	    unless ( $success ) {
		foreach $inc (@inc) {
		    ( $success = 1 ) if $inc eq $chk;
		}
	    }
	}
    }
    return 0 unless $success;
    
    # get interesting information if present
    @activity = ( $lineBuffer =~ /<activity[^<>]*>\s*([^<>]*?)\s*<\/activity/g );
    @locale = ( $lineBuffer =~ /<locale[^<>]*>\s*([^<>]*?)\s*<\/locale/g );
    $max = ( @activity > @locale ) ? @activity : @locale;
    $max = ( $max > $maxDesc ) ? $maxDesc : $max;
    $info = ' ';
    for ( $iDesc = 0; $iDesc < $max; $iDesc++ ) {
	$info .= ( defined( $locale[$iDesc] ) ? $locale[$iDesc] : '?' );
	$info .= ' - ' . ( defined( $activity[$iDesc] ) ? $activity[$iDesc] : '?' ) . '; ';
    }

    # discard <teiHeader> section from buffer and return
    $lineBuffer =~ s/^.*\/teiHeader>//;
    return $info;
}


# set up CP names & corresponding display colours
sub getHeaderInfo {
    my ( $iAnon, $iColour, $key, $value );

    @colours = ( 'Gray', 'Maroon', 'Red', 'Purple', 'Fuchsia', 
		 'Green', 'Lime', 'Olive', 'Yellow', 'Navy', 'Blue', 'Teal', 'Aqua' );

    # get whole <teiHeader> section
    while ( $lineBuffer .= <STDIN> ) {
	chomp( $lineBuffer );
	last if ( $lineBuffer =~ /\/teiHeader>/ );
    }

    # get CP names from <person> info if present, distinguishing between 'None' entries
    $iAnon = 0;
    $iColour = 0;
    %people = ( $lineBuffer =~ /<person[^<>]*?id=\"([^\"]*)\"[^<>]*>(.*?)<\/person/g );
    while (($key, $value) = each %people) {
	if ( ( $value =~ s/^.*<name>(\w+)<\/name.*$/$1/ ) and ( $value ne 'None' ) ) {
	    $people{ $key } = $value;
	}
	else {
	    $people{ $key } = "Anon " . ++$iAnon;
	}
	$whoColours{ $key } = $colours[($iColour++) % @colours];
    }

    # set up names & colours for global IDs
    $people{ 'PS000' } = 'Unknown';
    $people{ 'PS001' } = 'Unknowns';
    $whoColours{ 'PS000' } = $whoColours{ 'PS001' } = 'White';

    # discard <teiHeader> section from buffer and return
    $lineBuffer =~ s/^.*\/teiHeader>//;
}


# process a sentence SGML -> HTML suitable for display
sub noTags {
    my ( $notags ) = @_;
    my $l = '<em>&lt;';   # MUST contain > chars
    my $r = '&gt;</em>';  # MUST contain > chars
    my $odd = 1;
    $startTag = '<b><u>';
    $endTag = '</u></b>';

    ( $ptrColour = 0 ) unless defined $ptrColour;

    for ( $notags ) {

	# convert SGML entities
	s/\&mdash\;/\&\#8212\;/g;
	s/\&[be]quo\;/\&quot\;/g;

	# remove tags we don't want
	s/<w\s+\w{3}(-\w{3})?>//go;
	s/<c\s+PU\w>//go;
	s/<\/u>//go;
	s/<\/div>//go;
	s/<(\/)?stext[^>]*>//go;
	s/<bncDoc[^>]*>//go;
	s/<(\/)?align>//go;
	s/<loc[^>]*>//go;

	# make those we do want visible
	s/<([^>]*?)>/$l$1$r/go;

	# replace <trunc> sections by strikeout
	s/($l)trunc$r/<s>/go;
	s/($l)\/trunc$r/<\/s>/go;

	# make <sic> sections easier to read
	s/($l)sic[^>]*$r//go;
	s/($l)\/sic$r/<em>\[sic\]<\/em>/go;

	# make anonymisation gaps easier to read
	s/($l)gap\s[^>]*?desc=\"([^>]*?)\".*?$r/<em>\[$2\]<\/em>/go;
	s/($l)gap\s[^>]*?desc=(\w*)[^>]*?$r/<em>\[$2\]<\/em>/go;

	# and same for events
	s/($l)event\s[^>]*?desc=(\")?([^>\"]*)(\")?/$1$3/go;
	s/($l)vocal\s[^>]*?desc=(\")?([^>\"]*)(\")?/$1$3/go;

	# insert name for unclear utterances
	if ( /($l)unclear\s[^>]*?who=(\w+)/o ) {
	    $name = ( defined( $people{$2} ) ? $people{$2} : $2 );
	    s/(($l)unclear\s[^>]*?)who=(\w+)/$1 \[$name\]/go;
	}

	# replace <div>, <bncDoc> boundaries
	s/($l)div[^>]*($r)/<\/td><\/tr><tr><td colspan=3>----------------- New Conversation -----------------/o;
	s/($l)\/bncDoc($r)/<\/td><\/tr><tr><td colspan=3>----------------- End -----------------/o;
	s/($l)\/bncDoc($r)//go;

    }

    # colour-coding for overlapping <ptr> sections
    while ( $notags =~ /ptr\s+target=(\w+)$r/o ) {
	unless ( defined $ptrColourTags{$1} ) {
	    if ( $odd ) {
		$ptrColourTags{$1} = "<font COLOR=$colours[($ptrColour++) % @colours]>";
	    }
	    else {
		$ptrColourTags{$1} = '</font>';
	    }
	    $odd = !$odd;
	}
	$notags =~ s/($l)ptr\s+target=(\w+)$r/$ptrColourTags{$2}/o;
    }

    # table formatting for u/s tags
    $ended = 0 unless defined( $ended );
    for ( $notags ) {

	$newEnded = ( /($l)u\s+who=(\w+)\s*$r\s*($endMark\s*)?$/o );

	s/(($startMark)?)($l)u\s+who=(\w+)\s*$r\s*(?=($endMark\s*)?$|($l)s\s)/<\/td><\/tr><tr><td bgcolor=$whoColours{$4}><em>$people{$4}<\/em><\/td><td>$1/go;
	s/(($startMark)?)($l)u\s+who=(\w+)\s*$r\s*/"<\/td><\/tr><tr><td bgcolor=$whoColours{$4}><em>$people{$4}<\/em><\/td>" . ( $sentNum ? "<td><\/td>" : "" ) . "<td>$1"/goe;

	s/(?<!><td>|$startMark)(($startMark)?)($l)s\s+n=\"(\d+)\"\s*$r/"<\/td><\/tr><tr><td><\/td>" . ( $sentNum ? "<td><font SIZE=-2>$4<\/font><\/td>" : "" ) . "<td>$1"/goe unless $ended;
	s/(($startMark)?)($l)s\s+n=\"(\d+)\"\s*$r/( $sentNum ? "<font SIZE=-2>$4<\/font><\/td><td>$1" : "$1" )/goe;

    }
    $ended = $newEnded;

    # finally, replace __B/E__ markers
    $notags =~ s/$startMark/$startTag/go;
    $notags =~ s/$endMark/$endTag/go;

    return $notags;
}


# convert search string to reg-exp suitable for searching SGML
sub stringToRegExp {
    my ( $arg, $caseSens, $punctSens ) = @_;

    $sp = ( $punctSens ? '\s*' : '(?:<c PU\w>\W)?\s*' );
    $cs = ( $caseSens ? '' : '(?i)' );
    $br = $uttSep;

    # translate wildcards to regexp
    for ( $arg ) {

	# replace whitespace by tabs to make it easier to replace uniquely later
	s/\s+/\t/go;

	# convert wildcards to regexp equivalents
	s/\*/\\w\*/go;        # * -> \w*
	s/\+/\\w\+/go;        # + -> \w+
	s/(\S)\?/$1\\w/go;    # ? -> \w  (if not preceded by whitespace)
	s/\?(\S)/\\w$1/go;    # ? -> \w  (if not followed by whitespace)
	s/\[/\(\?:/go;        # [ -> (?:
	s/\]/\)\?/go;         # ] -> )?
	s/(\S+),/$1\t\.\*\?/go;  # , -> .*? (if not preceded by whitespace)

	# first deal with PoSs (currently the only <> tags in $arg)
	s/<([^<>]+)>/<w (?:\\w{3}-)?$1(?:-\\w{3})?>(\\w+)\\b/go;

	# deal with sentence/turn breaks (don't allow .* if ^, $ adjacent)
        s/(\$\t)(\|)(\t\^)/$1($br)$3/go;
        s/(^|\t)(\|)(\t\^)/$1(.*?$br)$3/go;
        s/(\$\t)(\|)(\t|$)/$1($br.*?)$3/go;
        s/(^|\t)(\|)(\t|$)/$1(.*?$br.*?)$3/go;

	# add required tags to ^, $, keeping (^/$) only if actually at beginning/end
	s/(^)(\^)(\t|$)/($2)\\s*(?:<[^<>]*>\\s*)*$3/go;
	s/(\t)(\^)(\t|$)/$1\\s*(?:<[^<>]*>\\s*)*$3/go;
        s/(^|\t)(\$)($)/$1(?:<c PU\\w>\\W)?\\s*(?:<[^<>]*>\\s*)*($2)$3/go;
        s/(^|\t)(\$)(\t)/$1(?:<c PU\\w>\\W)?\\s*(?:<[^<>]*>\\s*)*$3/go;

	# add PoS tag to punctuation marks
	s/(^|\t)([,.?!;:()])(\t|$)/$1<c PU\\w>(\\$2)$3/go;

        # add PoS tag to words, inside optional brackets if necessary
  	s/(^|\t)(\(\?:)([^.<\t]+)(\)\?)(\t|$)/$1$2<w \\w{3}(?:-\\w{3})?>($cs$3)\\b$4$5/go;
  	s/(^|\t)(\(\?:)([^.<\t]+)(\)\?)(\t|$)/$1$2<w \\w{3}(?:-\\w{3})?>($cs$3)\\b$4$5/go;
  	s/(^|\t)([^.<\t]+)(\t|$)/$1<w \\w{3}(?:-\\w{3})?>($cs$2)\\b$3/go;
  	s/(^|\t)([^.<\t]+)(\t|$)/$1<w \\w{3}(?:-\\w{3})?>($cs$2)\\b$3/go;

	# and replace tabs with required separator
	s/(^|\t)(\.\*\?)(\t|$)/$2/go;
	s/\t/$sp/go;

    }
    return $arg;

}


sub getSentence {

    while ( <STDIN> ) {
  	chomp( $lineBuffer .= $_ );

    	# beginning of next sentence? return complete one
  	if ( $lineBuffer =~ /$uttSep.*?$uttSep/o ) {
  	    $lineBuffer =~ s/^(.*?$uttSep.*?)($uttSep)/$2/o;
  	    return $1;
  	}
    }
    
    # end of file? return any dregs
    $lineBuffer =~ s/^(.*)$//o;
    return $1;

}


# ensure package returns 1 to keep Perl happy
1;
